/* CHAOS2 - sample code for a basic chaortic synthesis instrument
    This uses the Lorenz attractor (x map only):
 
        dx/dt = a(y-x)
        dy/dt = bx - y - zx
        dz/dt = xy -cz
 
        a = 10.0
        b = 28.0
        c = 8.0/3.0
 
    simple chaos instrument

   p0 = output start time
   p1 = duration
   p2 = amplitude multiplier
   p3 = delta value (typically 0.01)
   p4 = pan value (0 - 1)


*/
#include <stdio.h>
#include <stdlib.h>
#include <ugens.h>
#include <Ougens.h>
#include "CHAOS2.h"			  // declarations for this instrument class
#include <rt.h>
#include <rtdefs.h>


// Construct an instance of this instrument and initialize some variables.
// Using an underbar as the first character of a data member is a nice
// convention to follow, but it's not necessary, of course.  It helps you
// to see at a glance whether you're looking at a local variable or a
// data member.

CHAOS2::CHAOS2()
{
}


// Destruct an instance of this instrument, freeing any memory you've allocated.

CHAOS2::~CHAOS2()
{
}


// Called by the scheduler to initialize the instrument. Things done here:
//   - read, store and check pfields
//   - set output file (or bus) pointer
//   - init instrument-specific things
// If there's an error here (like invalid pfields), call and return die() to 
// report the error.  If you just want to warn the user and keep going,
// call warn() or rterror() with a message.

int CHAOS2::init(double p[], int n_args)
{
	// Tell scheduler when to start this inst.  If rtsetoutput returns -1 to
	// indicate an error, then return DONT_SCHEDULE.

	if (rtsetoutput(p[0], p[1], this) == -1)
		return DONT_SCHEDULE;

    amp = p[2];
    doupdatecheck = 0;

    x = 1.0;
    y = 1.0;
    z = 1.0;
    delta = p[3];

    pan = p[3];

	return nSamps();
}


// For non-interactive (script-driven) sessions, the constructor and init()
// for every instrument in the script are called before any of them runs.
// By contrast, configure() is called right before the instrument begins
// playing.  If we were to allocate memory at init time, then all notes in
// the score would allocate memory then, resulting in a potentially excessive
// memory footprint.  So this is the place to allocate any substantial amounts
// of memory you might be using.

int CHAOS2::configure()
{
	return 0;	// IMPORTANT: Return 0 on success, and -1 on failure.
}


// Called at the control rate to update parameters like amplitude, pan, etc.

void CHAOS2::doupdate()
{
	// The Instrument base class update() function fills the <p> array with
	// the current values of all pfields.  There is a way to limit the values
	// updated to certain pfields.  For more about this, read
	// src/rtcmix/Instrument.h.

	double p[4];
	update(p, 4);

    amp = p[2];

    delta = p[3];

    pan = p[4];
}


// Called by the scheduler for every time slice in which this instrument
// should run.  This is where the real work of the instrument is done.

int CHAOS2::run()
{
	// framesToRun() gives the number of sample frames -- 1 sample for each
	// channel -- that we have to write during this scheduler time slice.
	// Each loop iteration outputs one sample frame.
    
    float currentsample;
    float xnew, ynew, znew, dx, dy, dz;

	for (int i = 0; i < framesToRun(); i++) {
        if (--doupdatecheck <= 0) {
            doupdate();
            doupdatecheck = getSkip();
        }

		float out[2];		// Space for only 2 output chans!

        // here is the attractor equation!
        dx = 10.0 * (y-x);
        dy = x * (28.0-z) - y;
        dz = x*y - (8.0/3.0)*z;
        xnew = x + delta*dx;
        ynew = y + delta*dy;
        znew = z + delta*dz;
        
        x = xnew;
        y = ynew;
        z = znew;
    
        currentsample = (x/30.0) * amp;

        out[1] = currentsample * pan;
        out[0] = currentsample * (1.0 - pan);

        
		// Write this sample frame to the output buffer.

		rtaddout(out);

		// Increment the count of sample frames this instrument has written.

		increment();
	}

	// Return the number of frames we processed.

	return framesToRun();
}


// The scheduler calls this to create an instance of this instrument
// and to set up the bus-routing fields in the base Instrument class.
// This happens for every "note" in a score.

Instrument *makeCHAOS2()
{
	CHAOS2 *inst = new CHAOS2();
	inst->set_bus_config("CHAOS2");

	return inst;
}


// The rtprofile introduces this instrument to the RTcmix core, and
// associates a script function name (in quotes below) with the instrument.
// This is the name the instrument goes by in a script.

void rtprofile()
{
	RT_INTRO("CHAOS2", makeCHAOS2);
}


